Package: bayesRecon
Type: Package
Date: 2026-04-16
Title: Probabilistic Reconciliation via Conditioning
Version: 1.0.1
Authors@R: c(person(given = "Dario",
             family = "Azzimonti",
             role = c("aut","cre"),
             email = "dario.azzimonti@gmail.com",
             comment = c(ORCID = "0000-0001-5080-3061")),
       person(given = "Lorenzo",
             family = "Zambon",
             role = c("aut"),
             email = "lorenzo.zambon@idsia.ch",
             comment = c(ORCID = "0000-0002-8939-993X")),
       person(given = "Stefano",
             family = "Damato",
             role = c("aut"),
             email = "stefano.damato@idsia.ch",
             comment = c(ORCID = "0009-0001-0225-3736")),
      person(given = "Nicolò",
             family = "Rubattu",
             role = c("aut"),
             email = "nicolo.rubattu@idsia.ch",
             comment = c(ORCID = "0000-0002-2703-1005")),
      person(given = "Giorgio",
             family = "Corani",
             role = c("aut"),
             email = "giorgio.corani@idsia.ch",
             comment = c(ORCID = "0000-0002-1541-8384")))
Maintainer: Dario Azzimonti <dario.azzimonti@gmail.com>
Description: Provides methods for probabilistic reconciliation of hierarchical forecasts of time series. 
             The available methods include analytical Gaussian reconciliation (Corani et al., 2021) 
             <doi:10.1007/978-3-030-67664-3_13>, 
             MCMC reconciliation of count time series (Corani et al., 2024) 
             <doi:10.1016/j.ijforecast.2023.04.003>, 
             Bottom-Up Importance Sampling (Zambon et al., 2024) 
             <doi:10.1007/s11222-023-10343-y>,
             methods for the reconciliation of mixed hierarchies (Mix-Cond and TD-cond) (Zambon et al., 2024) 
             <https://proceedings.mlr.press/v244/zambon24a.html>,
             analytical reconciliation with Bayesian treatment of the covariance matrix (Carrara et al., 2025)
             <doi: 10.48550/arXiv.2506.19554>.
License: LGPL (>= 3)
Depends: R (>= 4.1.0)
Imports: stats, utils, lpSolve (>= 5.6.18), nloptr
Encoding: UTF-8
LazyData: true
RoxygenNote: 7.3.3
Suggests: knitr, rmarkdown, forecast, glarma, scoringRules, ggplot2,
        testthat (>= 3.0.0)
Config/testthat/edition: 3
VignetteBuilder: knitr
URL: https://github.com/IDSIA/bayesRecon,
        https://idsia.github.io/bayesRecon/
BugReports: https://github.com/IDSIA/bayesRecon/issues
NeedsCompilation: no
Packaged: 2026-04-16 09:54:44 UTC; dario.azzimonti
Author: Dario Azzimonti [aut, cre] (ORCID:
    <https://orcid.org/0000-0001-5080-3061>),
  Lorenzo Zambon [aut] (ORCID: <https://orcid.org/0000-0002-8939-993X>),
  Stefano Damato [aut] (ORCID: <https://orcid.org/0009-0001-0225-3736>),
  Nicolò Rubattu [aut] (ORCID: <https://orcid.org/0000-0002-2703-1005>),
  Giorgio Corani [aut] (ORCID: <https://orcid.org/0000-0002-1541-8384>)
Repository: CRAN
Date/Publication: 2026-04-16 10:32:05 UTC
Built: R 4.6.0; ; 2026-04-26 12:40:57 UTC; unix
