# Generated by roxygen2: do not edit by hand

export(.check_hierarchical)
export(.compute_naive_cov)
export(.core_reconc_BUIS)
export(.core_reconc_MixCond)
export(.core_reconc_TDcond)
export(.core_reconc_t)
export(.get_Au)
export(.get_HG)
export(.lowest_lev)
export(PMF_get_mean)
export(PMF_get_quantile)
export(PMF_get_var)
export(PMF_sample)
export(PMF_summary)
export(get_reconc_matrices)
export(multi_log_score_optimization)
export(reconc_BUIS)
export(reconc_MCMC)
export(reconc_MixCond)
export(reconc_TDcond)
export(reconc_gaussian)
export(reconc_t)
export(schaferStrimmer_cov)
export(temporal_aggregation)
import(nloptr)
